import logging
import os 
import numpy as np 
from .abcBackend import Backend as BackendBaseClass
import multiprocessing

import sys
thisFilePath = os.path.realpath(__file__)
pythonDir = os.path.dirname(os.path.dirname(thisFilePath))
sys.path.append(pythonDir)
from Simulator import simulatorQICS_exact as fSimulator  # fortran



f2pyFile = '/hidePath/fortranCircuit/fCircuit.cpython-39-x86_64-linux-gnu.so'
sys.path.append( os.path.dirname(f2pyFile) ) 
import fCircuit






# def _nCoreFunc(args):
#     self = args[0]
#     sim = args[1]
#     para = args[2]
#     res = sim.runCircuit(F=para['F'], I=para['I'])
#     return self._reformeSingleReulst(res)


# class Backend(BackendBaseClass):
#     """
#     A simulator written by fortran. Consider speed only.

#     Args:
#         circuit (Circuit): _description_
#         workDir (str, None): _description_. Defaults to None.
#         nCore (int,1): for parallel calculation.

#     """    

    
#     # """
#     # A simulator written by fortran. Consider speed only.

#     # Args:
#     #     BackendBaseClass (_type_): _description_
#     # """    


#     def runJob_noParameter(self)->dict:
#         sim = self._getSim()
#         res = sim.runCircuit()
#         return self._reformeSingleReulst(res)

#     def runJob_withParameters(self,paramList)->list:
#         sim = self._getSim()
#         res = sim.runCircuit() 
#         nCore = self.kw.get('nCore',1)
#         if nCore > 1:
#             with multiprocessing.Pool(self.nCore) as p:
#                 res = p.map( _nCoreFunc , [[self,sim,para] for para in paramList]   )
#                 return res
#         else:
#             return [ sim.runCircuit(F=para['F'], I=para['I']) for para in paramList]


#     def _reformeSingleReulst(self,res)->dict:
#         """
#         transform result from [old-version-simulator] into standard one 

#         Args:
#             res (dict): {0:p0,1:p1,...}

#         Returns:
#             dict: only return Probabilities for measured qubits
#         """   
#         def getbitFiled(n,L):
#             result = np.binary_repr(n,L)
#             r = [int(digit) for digit in result]
#             r.reverse()
#             return r 

#         qList = self.circuit.getMeasurementList()
#         l = [] 
#         for q in qList:
#             if q not in l:
#                 l.append(q) 
#         l.sort() 
#         L = self.circuit.getQN()
#         R = {}
#         for i in res:
#             # print(i,L,type(i),type(L),999)
#             bitField = getbitFiled(n=int(i),L=L)
#             R_key = "".join([str(bitField[q]) for q in qList])
#             if res[i] > 1e-14:
#                 if R_key not in R:
#                     R[R_key] = 0 
#                 R[R_key] += res[i]
#         return R         


#     def _getSim(self):
#         '''
#         '''
#         if hasattr(self,'SIM'):
#             sim = self.SIM 
#         else:
#             self.SIM = fSimulator(soFile = os.path.join(self.workDir,'resource.so'))
#             sim = self.SIM
#             nCore = self.kw.get('nCore',None)
#             if nCore is not None and nCore > 1:
#                 sim.initCalc() 
#         return sim 
        
        






























class findSimulator:
    @staticmethod
    def set_simulator_fileName():
        return "SIMULATOR"

    @staticmethod
    def get_dirName_of_Simulator(isqcDIR):
        storeDIR = os.path.join(isqcDIR, "nix", "store")
        messList = os.listdir(storeDIR)
        for l in messList:
            if "simulator" in l:
                return l.strip()
        logging.error("cannot find simulator dir")

    @staticmethod
    def get_isQ_dir():
        isqcBIN = os.popen("which isqc").read().strip()
        isqcDIR = os.path.dirname(isqcBIN)
        return isqcDIR

    @classmethod
    def get_simulatorBIN_path(cls):
        return os.path.join(cls.get_isQ_dir(), cls.set_simulator_fileName())

    @classmethod
    def makeSimulatorBIN(cls):
        # will automatically find dir of isqc, and make a SIMULATOR with appropreate chmod
        simBinFile = cls.get_simulatorBIN_path()
        if os.path.exists(simBinFile):
            return
        isQdir = cls.get_isQ_dir()
        messName = cls.get_dirName_of_Simulator(isQdir)
        replaceKeyWord = f"/nix/store/{messName}/bin/simulator"
        isqcFile = os.path.join(isQdir, "isqc")
        with open(isqcFile, "r") as f:
            isqcContent = f.read()
        slides = isqcContent.split(" ")
        slides[-2] = f"/nix/store/{messName}/bin/simulator"
        simBINcontent = " ".join(slides)
        with open(simBinFile, "w") as f:
            f.write(simBINcontent)
        os.chmod(simBinFile, 0o555)



class SIMULATOR(object):
    # base class

    def __init__(self, soFile):
        #         self.so = soFile
        self.so = os.path.abspath(soFile)
        self.fileDir = os.path.dirname(self.so)
        self.SimulatorBIN = findSimulator.get_simulatorBIN_path()
        findSimulator.makeSimulatorBIN()

    def getsoDir(self):
        return os.path.dirname(self.so)

    def getSoFileName(self):
        return os.path.basename(self.so)

    def get_qcis_SO_fileName(self):
        return self.getSoFileName().replace(".so", "_QCIS_TMP_.so")
    

    def get_SIMULATOR(self):
        return self.SimulatorBIN

    def get_isQPath(self):
        return self.so.split(".so")[0] + ".isq"

    @staticmethod
    def bitfield(n: int, L: int):  # -> list[int]
        result = np.binary_repr(n, L)
        r = [int(digit) for digit in result]  # [2:] to chop off the "0b" part
        return r

    @staticmethod
    def calculate(func, res):
        # can calculate sum_i <i|A|i> = sum_i p_i * a_i ,
        # func(bitField) -> float,  bitField is in a reversed order (lower index first)
        r = 0
        for b in res:
            bitField = [int(i) for i in b]
            bitField.reverse()
            #             print(res,bitField,666)
            r += func(bitField) * res[b]
        return r







class Circuit_FsimulatorInterface():

    def __init__(self,nq,**kw):
        self.nq = nq
        self.iParaList = []
        self.fParaList = []  
        self.iParaLen = 4 
        self.fParaLen = 1 

    def append(self,iPara=[],fPara=[]):
        IP = iPara + [0] * (self.iParaLen - len(iPara)) 
        FP = fPara + [0.]* (self.fParaLen - len(fPara)) 
        self.iParaList.append(IP) 
        self.fParaList.append(FP) 
        return self 

    def X(self,i):return self.append(iPara=[0,i])

    def Y(self,i):return self.append(iPara=[1,i])

    def Z(self,i):return self.append(iPara=[2,i])

    def H(self,i):return self.append(iPara=[3,i])

    def S(self,i):return self.append(iPara=[4,i])

    def T(self,i):return self.append(iPara=[5,i])

    def Rx(self,i,theta): return self.append(iPara=[6,i],fPara=[theta])

    def Ry(self,i,theta): return self.append(iPara=[7,i],fPara=[theta])

    def Rz(self,i,theta): return self.append(iPara=[8,i],fPara=[theta])

    def RX(self,i,theta): return self.Rx(i,theta)

    def RY(self,i,theta): return self.Ry(i,theta)

    def RZ(self,i,theta): return self.Rz(i,theta) 

    def X2P(self,i):return self.Rx(i,np.pi/2)

    def X2M(self,i):return self.Rx(i,-np.pi/2)

    def Y2P(self,i):return self.Ry(i,np.pi/2)

    def Y2M(self,i):return self.Ry(i,-np.pi/2)

    def SD(self,i):return self.append(iPara=[9,i])

    def TD(self,i):return self.append(iPara=[10,i])

    def CNOT(self,i,j):return self.append(iPara=[100,i,j])

    def CZ(self,i,j):return self.append(iPara=[101,i,j])

    def calc_state(self):
        IP = np.array(self.iParaList).reshape(-1)
        FP = np.array(self.fParaList).reshape(-1) 
        d = 2**self.nq 
        nGates = len(self.iParaList) 
        # print("d=",d) 
        # print("nGates=",nGates)
        # print('IP=',IP)
        # print('FP=',FP)
        r = fCircuit.simulatecircuit(d,nGates,IP,FP) 
        return r 

    def MeasureAll(self):
        vec = self.calc_state() 
        res = {} 
        for i in range(len(vec)):
            res[i] = np.real( vec[i] * np.conjugate(vec[i]) )
        return res 


class CircuitSimulatorByPharseQCIS:
    #  , ,,  'CY', 'CX', 'M', 'RXY'

    def pharseAll(self):
        lines = self.qcisStr.split("\n")
        for l in lines:
            l = l.strip()
            if not l:
                continue
            self.pharseOneLine(l.split(" "))

    def pharseOneLine(self, segs):
        gateName = segs[0]
        if gateName in [
            "H",
            "X",
            "Y",
            "Z",
            "S",
            "T",
            "SD",
            "TD",
            "X2M",
            "X2P",
            "Y2M",
            "Y2P",
        ]:
            getattr(self.c, gateName)(int(segs[1][1:]))
        elif gateName in ["RZ", "RX", "RY"]:
            getattr(self.c, gateName)(int(segs[1][1:]), float(segs[2]))
        elif gateName in ["CZ", "CNOT"]:
            getattr(self.c, gateName)(int(segs[1][1:]), int(segs[2][1:]))
        elif gateName == "M":
            pass
        else:
            logging.error(f"Undefined gatenName = {gateName}")

    @staticmethod
    def check(line_data):
        qdic = {}
        qnum = 0
        for idx, line in enumerate(line_data):
            line = line.strip()
            if not line:
                continue
            strArr = line.split(" ")

            if strArr[1] not in qdic:
                qdic[strArr[1]] = qnum
                qnum += 1

            if strArr[0] in ["CZ", "CY", "CX", "CNOT"]:
                if strArr[2] not in qdic:
                    qdic[strArr[2]] = qnum
                    qnum += 1
        return qnum, qdic

    def __init__(self, qcisStr, **kw):
        self.kw = kw
        self.qnum, self.qdic = self.check(qcisStr.split("\n"))
        self.c = Circuit_FsimulatorInterface(self.qnum)
        self.qcisStr = qcisStr

    def Measure(self):
        self.pharseAll()
        res = self.c.MeasureAll()
        return res


def _nCoreFunc(args):
    self = args[0]
    para = args[1]
    qcisCode = self.getQCIScode(F=para['F'],I=para['I'])
    phCir = CircuitSimulatorByPharseQCIS( qcisCode )
    res = phCir.Measure()
    return self._reformeSingleReulst(res)


class Backend(BackendBaseClass,SIMULATOR):
    """
    A simulator written by fortran. Consider speed only.

    Args:
        circuit (Circuit): _description_
        workDir (str, None): _description_. Defaults to None.
        nCore (int,1): for parallel calculation.

    """    

    
    # """
    # A simulator written by fortran. Consider speed only.

    # Args:
    #     BackendBaseClass (_type_): _description_
    # """   
    # 

    def __init__(self,circuit,workDir=None,**kw):
        BackendBaseClass.__init__(self,circuit,workDir=workDir,**kw)
        soFile = os.path.join( self.workDir, 'resource.so' )
        SIMULATOR.__init__(self,soFile=soFile)

    def make_qcis_resource(self,nF,nI)->str:
        '''create files without check exist'''
        fakeSO = os.path.join(self.getsoDir(), self.get_qcis_SO_fileName())
        passF = "".join([f" -d {f}" for f in [0.0]*nF])
        passI = "".join([f" -i {i}" for i in [0]*nI])
        cmd = f"isqc compile {self.get_isQPath()} --target qcis {passI} {passF} -o {fakeSO}"
        cwd = os.getcwd()
        os.chdir(self.workDir)
        os.system(cmd)
        os.chdir(cwd) 
        return fakeSO
    
    def make_qcis_if_need(self,nF,nI)->str:
        if self.isCreateIsQNeeded:
            self.make_qcis_resource(nF,nI)
 
    def getQCIScode(self, F=[], I=[], **kw):
        simulatorBIN = self.get_SIMULATOR()
        self.FAKE_SO = os.path.join(self.getsoDir(), self.get_qcis_SO_fileName())
        passF = "".join([f" -d {f}" for f in F])
        passI = "".join([f" -i {i}" for i in I])
        # print("here555",self.FAKE_SO);raise 
        cmd = (
            f"{simulatorBIN} {self.FAKE_SO} --qcisgen -e __isq__entry {passI} {passF} "
        )
        cwd = os.getcwd()
        os.chdir(self.workDir)
        QCIS = os.popen(cmd).read()
        os.chdir(cwd) 
        return QCIS

    def runJob_noParameter(self)->dict:
        self.make_qcis_if_need(0,0) 
        qcisCode = self.getQCIScode() 
        phCir = CircuitSimulatorByPharseQCIS( qcisCode)
        res = phCir.Measure()
        return self._reformeSingleReulst(res)

    def runJob_withParameters(self,paramList)->list:
        if len(paramList)==0:return []
        nF = len(paramList[0]['F'])
        nI = len(paramList[0]['I'])
        self.make_qcis_if_need(nF=nF,nI=nI) 
        nCore = self.kw.get('nCore',1)
        if nCore > 1:
            qcisCode = self.getQCIScode() 
            with multiprocessing.Pool(self.nCore) as p:
                res = p.map( _nCoreFunc , [[self,para] for para in paramList]   )
                return res
        else:
            RETURN = [] 
            for para in paramList:
                qcisCode = self.getQCIScode(F=para['F'],I=para['I']) 
                phCir = CircuitSimulatorByPharseQCIS( qcisCode )
                RETURN.append( self._reformeSingleReulst( phCir.Measure() ) )
            return RETURN


    def _reformeSingleReulst(self,res)->dict:
        """
        transform result from [old-version-simulator] into standard one 

        Args:
            res (dict): {0:p0,1:p1,...}

        Returns:
            dict: only return Probabilities for measured qubits
        """   
        def getbitFiled(n,L):
            result = np.binary_repr(n,L)
            r = [int(digit) for digit in result]
            r.reverse()
            return r 

        qList = self.circuit.getMeasurementList()
        l = [] 
        for q in qList:
            if q not in l:
                l.append(q) 
        l.sort() 
        L = self.circuit.getQN()
        R = {}
        for i in res:
            # print(i,L,type(i),type(L),999)
            bitField = getbitFiled(n=int(i),L=L)
            R_key = "".join([str(bitField[q]) for q in qList])
            if res[i] > 1e-14:
                if R_key not in R:
                    R[R_key] = 0 
                R[R_key] += res[i]
        return R         


    def _getSim(self):
        '''
        '''
        if hasattr(self,'SIM'):
            sim = self.SIM 
        else:
            self.SIM = fSimulator(soFile = os.path.join(self.workDir,'resource.so'))
            sim = self.SIM
            nCore = self.kw.get('nCore',None)
            if nCore is not None and nCore > 1:
                sim.initCalc() 
        return sim 
        
        
