import logging
from .isQmaker import isQ_maker


class isqInputArgument:
    @staticmethod
    def getParaName():
        return "intPara", "floPara"

    def __init__(self, FI: str, id: int):
        """
        FI = 'F' for float (double), 'I' for int
        """
        self.FI = FI
        self.id = id

    def get_paraName_in_isq(self):
        # paraNames = self.getParaName()

        return self.getParaName()[{"F": 1, "I": 0}[self.FI]] + f"[{self.id}]"


class Circuit:
    """
    Essentially this class is used to maintain a pice of isQ code. No direct operations can be done on it.
    To interact with circuit, one must use a circuitDeployer

    [init] nq,isInputArg=False
           nq: number of qubit in total
           if isInputArg is True, isq contains argument from outside

    [fun] getQN()
          get nq

    [fun]

    """

    def __init__(self, nq: int, isInputArg=False):
        self.nq = nq
        self.isInputArg = isInputArg  # input arguments
        self.PREAMBLES_STORE = []
        self.CODE_VARIABLE_DEFINITION = []
        self.CODE_PROGRAM_MAIN_BODY = []
        self.CODE_MEARSUREMENT = []
        self.DEPLOYER_STORE = []
        self.cls = {
            "inputArgumentClass": isqInputArgument,
            "isQmaker": isQ_maker,
        }

        

    def getInputArg(self, FI: str, id: int):
        """
        FI = 'F' or 'I'
        get the <id>-th input argument
        """
        if self.isInputArg:
            return isqInputArgument(FI, id)
        else:
            logging.error(
                "circuit not support input argument, consider setting isInputArg=True"
            )

    def setMeasurement(self,qList):
        """set measurement on qubits

        Args:
            qList (list): int number represent the index of qbuit that will be measured. 
        """ 
        self.CODE_MEARSUREMENT += qList
        # for qi in qList:
        #     if qi in self.CODE_MEARSUREMENT:
        #         logging.error(f"measurement on qubit {qi} more than once.")
        #     else:
        #         self.CODE_MEARSUREMENT.append(qi) 
        # self.CODE_MEARSUREMENT.sort() 
        # 

    def getMeasurementList(self):
        return list(self.CODE_MEARSUREMENT)   




    def isInputParameter(self):
        return self.isInputArg

    def getQN(self):
        return self.nq

    # def get_isQ_maker(self):
    #     return self.cls['isQmaker']

    def defineVariable(self, FI, varName, value):
        Type = {"F": "double", "I": "int"}[FI]
        self.CODE_VARIABLE_DEFINITION.append(f"{Type} {varName} = {value};")
        return self

    def defineArray_autoSet(self, varName, data):
        self.CODE_VARIABLE_DEFINITION.append(
            isQ_maker.autoSetArray(varName=varName, data=data)
        )
        return self

    def registDeployer(self):
        self.DEPLOYER_STORE.append({})
        rid = len(self.DEPLOYER_STORE) - 1
        return rid, self.DEPLOYER_STORE[rid]

    def appendPreambleIfNotExist(self, preamble):
        if preamble not in self.PREAMBLES_STORE:
            self.PREAMBLES_STORE.append(preamble)

    def appendCode(self, codeBlock):
        self.CODE_PROGRAM_MAIN_BODY.append(codeBlock)
        return self

    def _makeCode(self):
        preamble = "\n".join(self.PREAMBLES_STORE)
        varDef = "\n".join(self.CODE_VARIABLE_DEFINITION)
        body = "\n".join(self.CODE_PROGRAM_MAIN_BODY)
        measure  = "\n".join( [ f"M ( Q[{q}] );" for q in self.CODE_MEARSUREMENT ] ) 
        inputArgStr = (
            " int {0} [], double {1} []".format(*isqInputArgument.getParaName())
            if self.isInputArg
            else ""
        )
        # code = f"""
        # { preamble }
        # qbit Q[{self.nq}];
        # procedure main( {inputArgStr} ) {{
        #     // ------ define variables --------
        #    {varDef}
        #     // ------ statements -----------
        #    {body} 
        # }}
        # """
        code = (
           f"{preamble} \n"
           f"qbit Q[{self.nq}];\n"
           f"procedure main( {inputArgStr} ) {{\n"
           f"// ------ define variables --------\n"
           f"{varDef}\n"
           "// ------ statements -----------\n"
           f"{body}\n"
           "// ------ measurement ------------\n"
           f"{measure}\n"
           "}" 
        )


        return code
