import numpy as np
from .abcDeployer import BaseDeployer
import logging

class Deployer(BaseDeployer):
    """
    target: lib.isQ.Physics.pauliHam.isq


    """

    def _Preambles(self):
        return ["import demoLib.isQ.Physics.pauliHam;"]

    def expHt(self, h, t, N):
        exist, varName = self.get_isqVarName_by_pyObject(h)
        varName_xi = varName + "_xi"
        varName_P = varName + "_P"
        if not exist:
            xiList = np.array(h.get_factor())
            PList = np.array(h.get_pauliOperators())
            self.circuit.defineArray_autoSet(varName=varName_xi, data=xiList)
            self.circuit.defineArray_autoSet(varName=varName_P, data=PList)
        varName_t = self.get_isqVarNameWithAppendIsQ(t)
        varName_N = self.get_isqVarNameWithAppendIsQ(N)
        varName_Q = "[" + ",".join([f"Q[{i}]" for i in self.qMap]) + "]"
        self.circuit.appendCode(
            f"deploy_expH_TS( {varName_xi} , {varName_P}, {varName_t}, {varName_Q}, {varName_N} ); "
        )
        return self 
    
    def gate4PauliMeasure(self,P):
        """_summary_

        Args:
            P (list): A list of integers to represent a pauli operator P. 
            To measure P directly on circuit, we need a (set of) conditional operator(s). This gate implement that.  

        Returns:
            self: 
        """        
        pTuple = tuple(P)
        exist, varName = self.get_isqVarName_by_pyObject(pTuple)
        varName_Q = self.getString_AllQubit()
        if not exist:
            self.defineArray_autoSet(varName=varName,data=P) 
        self.circuit.appendCode(
            f"deploy_PauliOperator_Measurement( {varName_Q}, {varName} ); "
        )
        return self 
    
    def gate4PauliMeasureOnHamiltonian(self,h,i):
        r"""For a given Hamiltonian 
            $ H = \sum_i \xi_i P_i $, 
        when we want to calculate \langle P_i \rangle, we need to deploy gate by `gate4PauliMeasure`. This sub automatically
        find the corresponding gate to deploy.   

        Args:
            h (pauli Hamiltonian): Hamiltonian object
            i (int): the i-th term in h
        """
        pass

